#! /usr/bin/perl -wT

package Twici ;

use strict ;

# todo: figure this stuff out
# use Exporter ;
# our @ISA = qw(Exporter) ;
# our @EXPORT = qw(handleView handleEdit) ; # why is this not necessary?

use POSIX qw(strftime) ;

use Fcntl ':flock' ;

use BSD::Resource ;
BEGIN
{
	my $hdlr = sub { my $sig = shift ; die( "resource limit exceeded; killed by a SIG$sig" ) ; } ;
	$SIG{"XCPU"} = $hdlr ;
	# there doesn't seem to be any way to catch the OOM signal
#	setrlimit( RLIMIT_CPU, 10, 11 ) ; # 10:11 sec
	setrlimit( RLIMIT_CPU, 20, 22 ) ; # 20:22 sec; just for now
	setrlimit( RLIMIT_VMEM, 104857600, 115343360 ) ; # 100:110 MB
}

### NOTES

# throughout, subroutines are called in the modern style, with parentheses and without
# ampersands (so they look like they do in real languages)

# constants are written inline where possible

# sets are handled as hashes, in which contained objects are keys mapping to undef

# todo: use references
# todo: use pod
# todo: in forms, make order of method + action consistent; action first

### CONFIGURATION

use constant WIKI_ROOT => "/srv/www/urchin.earth.li/users/twic/wiki/" ;
use constant WIKI_SRV => "http://urchin.earth.li/" ;
use constant WIKI_BASE => WIKI_SRV . "cgi-bin/twic/wiki/" ;
use constant WIKI_STATIC => WIKI_SRV . "~twic/wiki/" ;
use constant KEY_PAGES => ("FrontPage", "RecentChanges", "MessageOfTheDay") ;

### CONSTANTS

use constant WIKI_EXT => ".twici" ;
use constant BADURL_PAGE => "BadURL" ;
use constant BADPIC_PAGE => "BadPicture" ;
use constant NOTFOUND_PAGE => "PageNotFound" ;
use constant URL_RE => "([a-zA-Z][0-9a-zA-Z+\\-\\.]*):([^\\s>]+)" ;
my %LIST_ELEMS = ("-" => "ul", "#" => "ol") ; # perl sucks
use constant LAST_MODIFIED => "Last-Modified" ;
my %LOCK_HIER = ("<" => "-", ">" => ">", "-" => "-") ;

### HANDLERS

sub handleView
{
	validateRequestMethod( "GET" ) ;
	my $queryString = $ENV{"QUERY_STRING"} ;
	my %query = parse( $queryString ) ;
	my $page = validateWikiName( $query{"page"} ) ;
	my %hdrs = () ;
	if (!pageExists( $page ))
	{
		$page = NOTFOUND_PAGE ;
		$hdrs{"URI"} = viewUrl( NOTFOUND_PAGE ) ;
	}
	my $fh = openPage( $page, "<" ) ;
	my $mtime = mtime( $fh ) ;
	my $mtimeStr = timeToStr( $mtime ) ;
	$hdrs{"Last-Modified"} = $mtimeStr ;
	my $explodedPage = explodeWikiName( $page ) ;
	$hdrs{"Title"} = $explodedPage ;
	htmlHeader( $explodedPage, %hdrs ) ;
	renderPage( $page, $fh ) ;
	htmlFooter( $page ) ;
}

sub handleEdit
{
	validateRequestMethod( "GET" ) ;
	my $queryString = $ENV{"QUERY_STRING"} ;
	my %query = parse( $queryString ) ;
	my $page = validateWikiName( $query{"page"} ) ;
	my $explodedPage = explodeWikiName( $page ) ;
	htmlHeader( "Edit $explodedPage" ) ;
	my $action = updateUrl( $page ) ;
	print <<EOF ;
<form action="$action" method="POST">
<textarea name="text" rows="20" cols="80">
EOF
	if (pageExists( $page ))
	{
		my $fh = openPage( $page, "<" ) ;
		while (my $line = <$fh>) # todo: would io be more efficient if this was in block/slurp mode?
		{
			print $line ;
		}
		close( $fh ) ;
	}
	print <<EOF ;
</textarea>
<br>
<input type="submit">
<input type="reset">
</form>
EOF
	htmlFooter() ;
}

# todo: it would be nice and neat if view, edit and update all used blockmode. if it worked and was efficient, it should be done.

sub handleUpdate
{
	validateRequestMethod( "POST" ) ;
	my $queryString = $ENV{"QUERY_STRING"} ;
	my %query = parse( $queryString ) ;
	my $page = validateWikiName( $query{"page"} ) ;
	my $contentType = $ENV{"CONTENT_TYPE"} ;
	(defined( $contentType ) && (lc( $contentType ) eq "application/x-www-form-urlencoded")) || die( "undefined or incorrect content type" ) ;
	# todo: for god's sake use CGI.pm or CGI::Lite for this bit!
	# todo: support text/twici as content-type, and just dump it to $text
	my $in ;
	my $contentLength = $ENV{"CONTENT_LENGTH"} ;
	if ($contentLength)
	{
		read( STDIN, $in, $contentLength ) ;
	}
	else # todo: would this be more efficient (and would it work) in block mode?
	{
		my $linesep = $/ ;
		$/ = undef ;
		$in = <STDIN> ;
		$/ = $linesep ;
	}
	my %form = parse( $in ) ;
	my $text = $form{"text"} ;
	defined( $text ) || die( "text not defined in form: $in" ) ;
	$text = normaliseLines( $text ) ;
	my $fh = openPage( $page, ">") ;
	print $fh $text || die( "could not write to page file: $page ($!)" ) ;
	close( $fh ) ;
	my $explodedPage = explodeWikiName( $page ) ;
	htmlHeader( "Edited $explodedPage" ) ;
	my $linkHtml = wikiLinkHtml( $page ) ;
	my $editUrl = editUrl( $page ) ;
	print <<EOF ;
<p>You have edited $linkHtml; you should view it to check that your
changes are okay. Alternatively, you may <a href="$editUrl">edit</a> it
again if you like.</p>
EOF
	htmlFooter() ;
}

sub handleChanges
{
	validateRequestMethod( "GET" ) ;
	my $queryString = $ENV{"QUERY_STRING"} ;
	my %query = parse( $queryString ) ;
	my $days = $query{"days"} ;
	my $num = $query{"num"} ;
	htmlHeader( "Recent Changes" ) ;
	if (defined( $days ) || defined( $num ))
	{
		my $msg = "These are the " . ($num ? "last $num" : "") . " changes" . ($days ? " in the last $days days" : "") . "." ;
		print <<EOF ;
<p>$msg</p>
EOF
		my @changes = changes( $days, $num ) ;
		print "<ul>\n" ;
		foreach my $page (@changes)
		{
			my $linkHtml = wikiLinkHtml( $page ) ;
			my $mtime = pageMtime( $page ) ;
			my $mtimestr = timeToStr( $mtime ) ;
			print "<li>$linkHtml ($mtimestr)</li>\n" ;
		}
		print "</ul>\n" ;
	}
	else
	{
		my $action = changesUrl() ;
		print <<EOF ;
<p>This form can be used to query the wiki for a list of pages which have changed recently.</p>
<form action="$action" method="GET">
<table>
<tr><th>Days since the change</th><td><input type="text" name="days"></td></tr>
<tr><th>Number of changes since the change</th><td><input type="text" name="num"></td></tr>
</table>
<input type="submit">
<input type="reset">
</form>
<p>In both fields, blank or 0 means that the criterion is not applied.</p>
EOF
	}
	htmlFooter() ;
}

sub handleViews
{
	validateRequestMethod( "GET" ) ;
	my $queryString = $ENV{"QUERY_STRING"} ;
	my %query = parse( $queryString ) ;
	my $days = $query{"days"} ;
	my $num = $query{"num"} ;
	htmlHeader( "Recent Views" ) ;
	if (defined( $days ) || defined( $num ))
	{
		my $msg = "These are the " . ($num ? "last $num" : "") . " views" . ($days ? " in the last $days days" : "") . "." ;
		print <<EOF ;
<p>$msg</p>
EOF
		my @views = views( $days, $num ) ;
		print "<ul>\n" ;
		foreach my $page (@views)
		{
			my $linkHtml = wikiLinkHtml( $page ) ;
			my $atime = pageAtime( $page ) ;
			my $atimestr = timeToStr( $atime ) ;
			print "<li>$linkHtml ($atimestr)</li>\n" ;
		}
		print "</ul>\n" ;
	}
	else
	{
		my $action = viewsUrl() ;
		print <<EOF ;
<p>This form can be used to query the wiki for a list of pages which have been viewed recently.</p>
<form action="$action" method="GET">
<table>
<tr><th>Days since the view</th><td><input type="text" name="days"></td></tr>
<tr><th>Number of views since the view</th><td><input type="text" name="num"></td></tr>
</table>
<input type="submit">
<input type="reset">
</form>
<p>In both fields, blank or 0 means that the criterion is not applied.</p>
EOF
	}
	htmlFooter() ;
}

sub handleBacklinks
{
	validateRequestMethod( "GET" ) ;
	my $queryString = $ENV{"QUERY_STRING"} ;
	my %query = parse( $queryString ) ;
	my $page = validateWikiName( $query{"page"} ) ;
	my %backlinks = backlinks( $page ) ;
	my @backlinksLst = sort( keys( %backlinks ) ) ;
	my $pageUrl = viewUrl( $page ) ;
	my $explodedPage = explodeWikiName( $page ) ;
	htmlHeader( "Backlinks to $explodedPage" ) ;
	print <<EOF ;
<ul>
EOF
	foreach my $backlink (@backlinksLst)
	{
		my $backlinkHtml = wikiLinkHtml( $backlink ) ;
		print "<li>$backlinkHtml</li>\n" ;
	}
	print <<EOF ;
</ul>
EOF
	htmlFooter() ;
}

sub validateRequestMethod
{
	my $properMethod = shift ;
	$properMethod || die( "cannot validate request method due to idiot programmer" ) ;
	my $actualMethod = $ENV{"REQUEST_METHOD"} ;
	($actualMethod eq $properMethod) || die( "script must be accessed by $properMethod (not $actualMethod)" ) ;
}

### RENDERING

sub renderPage
{
	my $page = shift ;
	my $fh = shift ;
	my $oldSlash = $/ ;
	$/ = "" ; # paragraph mode
	while (my $block = <$fh>)
	{
		renderBlock( $block ) ;
	}
	close( $fh ) ;
	$/ = $oldSlash ;
}

sub renderBlock
{
	my $block = shift ;
	$block =~ /^(.*?)\s*$/s ;
	$block = $1 ;
	my $prefix = substr( $block, 0, 1 ) ;
	if (($prefix eq "-") || ($prefix eq "\#"))
	{
		renderList( $block ) ;
	}
	elsif ($prefix eq ":")
	{
		renderTaggedList( $block ) ;
	}
	elsif ($prefix eq "=")
	{
		renderHeading( $block ) ;
	}
	elsif ($prefix eq "[")
	{
		renderPicture( $block ) ;
	}
	elsif ($prefix eq "|")
	{
		renderTable( $block ) ;
	}
	elsif ($prefix eq "\$")
	{
		renderRaw( $block ) ;
	}
	# if adding a new non-text block, update isTextBlock
	else
	{
		renderParagraph( $block ) ;
	}
}

sub isTextBlock # tells if a block is text-type or not
{
	my $block = shift ;
	my $prefix = substr( $block, 0, 1 ) ;
	return (($prefix ne "[") && ($prefix ne "\$")) ;
}

sub renderParagraph
{
	my $para = shift ;
	print "<p>" ;
	renderText( $para ) ;
	print "</p>\n" ;
}

sub renderHeading
{
	my $hdg = shift ;
	$hdg =~ /^(=+)\s*(.*)$/s || die( "malformed heading: $hdg" ) ;
	my $prefix = $1 ;
	my $lev = length( $prefix ) + 1 ;
	$hdg = $2 ;
	print "<h$lev>" ;
	renderText( $hdg ) ;
	print "</h$lev>\n" ;
}

sub renderPicture
{
	my $pic = shift ;
	$pic =~ /^\[(.*?)\s*\]?$/s || die( "malformed picture: $pic" ) ;
	$pic = $1 ;
	$pic =~ /^(\S*)[ \t]*\n?(.*)$/s || die( "malformed picture: $pic" ) ;
	my $type ;
	my $body ;
	if ($2)
	{
		$type = $1 ;
		$body = $2 ;
	}
	else
	{
		$body = $1 ;
	}
	my $html = pictureHtml( $type, $body ) ;
	print "<center>$html</center>\n" ;
}

# body includes leading whitespace!
sub pictureHtml
{
	my $type = shift ;
	my $body = shift ;
	$type || ($type = "img") ;
	$type = lc( $type ) ;
	if ($type eq "img")
	{
		$body =~ /^(\S+)\s*(.*)$/ || die ( "malformed img body: $body" ) ;
		my $url = $1 ;
		my $alt = $2 ;
		my $altStr = $alt ? " alt=\"$alt\"" : "" ;
		return "<p><img src=\"$url\"$altStr></p>" ;
	}
	elsif ($type eq "ascii")
	{
		return "<table><tr><td><pre>\n$body\n</pre></td></tr></table>" ;
	}
	else
	{
		my $url = viewUrl( BADPIC_PAGE ) ;
		return "<pre>[<a href=\"$url\">@{[BADPIC_PAGE]}</a>]</pre>" ;
	}
}

sub renderList
{
	my $list = shift ;
	my @lines = split( "\n", $list ) ;
	my $item = "" ;
	my $stack = "" ;
	push( @lines, "-" ) ;
	foreach my $line (@lines)
	{
		my $key = substr( $line, 0, 1 ) ;
		if (($key eq "-") || ($key eq "#"))
		{
			if ($item ne "")
			{
				$stack = renderListItem( $item, $stack ) ;
			}
			$item = $line ;
		}
		else
		{
			$item .= "\n" . $line ;
		}
	}
	renderListPrefix( $stack, "" ) ;
}

sub renderListItem
{
	my $item = shift ;
	my $stack = shift ;
	$item =~ /^([\-\#]*)\s*(.*)$/s || die( "illegal list item: $item" ) ;
	my $pre = $1 ;
	$item = $2 ;
	renderListPrefix( $stack, $pre ) ;
	print "<li>" ;
	renderText( $item ) ;
	print "</li>\n" ;
	return $pre ;
}

sub renderListPrefix
{
	my $stack = shift ;
	my $pre = shift ;
	if ($pre ne $stack)
	{
		my ($root, $off, $on) = factorStrings( $stack, $pre ) ;
		while ($off ne "")
		{
			my $type = substr( $off, -1 ) ;
			$off = substr( $off, 0, -1 ) ;
			my $elem = $LIST_ELEMS{$type} ;
			print "</$elem>\n" ;
		}
		while ($on ne "")
		{
			my $type = substr( $on, 0, 1 ) ;
			$on = substr( $on, 1 ) ;
			my $elem = $LIST_ELEMS{$type} ;
			print "<$elem>\n" ;
		}
	}
}

sub renderTaggedList
{
	my $list = shift ;
	print "<dl compact>\n" ;
	my @lines = split( "\n", $list ) ;
	my $item = "" ;
	foreach my $line (@lines)
	{
		my $key = substr( $line, 0, 1 ) ;
		if ($key eq ":")
		{
			if ($item ne "")
			{
				renderTaggedListItem( $item ) ;
			}
			$item = $line ;
		}
		else
		{
			$item .= "\n" . $line ;
		}
	}
	renderTaggedListItem( $item ) ;
	print "</dl>\n" ;
}

sub renderTaggedListItem
{
	my $item = shift ;
	$item =~ /^:\s*(.*?)\s*:(.*)$/ || die( "illegal tagged list item: $item" ) ;
	my $tag = $1 ;
	$item = $2 ;
	print "<dt>" ;
	renderText( $tag ) ;
	print "</dt><dd>" ;
	renderText( $item ) ;
	print "</dd>\n" ;
}

sub renderTable
{
	my $tbl = shift ;
	print "<table>\n" ;
#	print "<table border>\n" ; # todo: decide whether borders or not
	my @lines = split( "\n", $tbl ) ;
	my $row = "" ;
	my $hdg = 1 ;
	foreach my $line (@lines)
	{
		if (substr( $line, 0, 1 ) eq "|")
		{
			if ($row ne "")
			{
				renderTableRow( $row, $hdg ) ;
				$hdg = 0 ;
			}
			$row = $line ;
		}
		else
		{
			$row .= "\n" . $line ;
		}
	}
	if ($row ne "")
	{
		renderTableRow( $row, $hdg ) ;
	}
	print "</table>\n" ;
}

sub renderTableRow
{
	my $row = shift ;
	my $hdg = shift ;
	$row =~ /^\|(.*)$/s || die( "invalid row: $row" ) ;
	$row = $1 ;
#	my @cells = split( "|", $row ) ; # aargh!
	my @cells = split( /\|/, $row ) ;
	my $tag = $hdg ? "th" : "td" ;
	print "<tr>" ;
	foreach my $cell (@cells)
	{
		print "<$tag>" ;
		$cell =~ /^\s*(.*?)\s*$/s || die( "impossible cell: $cell" ) ;
		$cell = $1 ;
		renderText( $cell ) ;
		print "</$tag>" ;
	}
	print "</tr>\n" ;
}

sub renderRaw
{
	my $raw = shift ;
	$raw =~ /^\$(\S*)\s+(.*)$/s ;
	my $type = $1 ;
	my $text = $2 ;
	if (length( $type ) > 0)
	{
		print "<!-- $type -->\n" ;
	}
	print "<pre>\n" ;
	print escape( $text ) ;
	print "\n</pre>\n" ;
}

# this is V2.0 of the text-level renderer
# it doesn't handle unwrapped URLs
sub renderText
{
	my $text = shift ;
	(length( $text ) > 0) || return ;
	$text =~ tr/ \t\n/ /s ;
	my @tokens = atomise( $text ) ;
	if ($tokens[0] eq " ")
	{
		shift @tokens ;
	}
	if ($tokens[-1] eq " ")
	{
		pop @tokens ;
	}
	my $em = 0 ;
	my $strong = 0 ;
	foreach my $tok (@tokens)
	{
		if ($tok eq "_")
		{
			if (!$em)
			{
				print "<em>" ;
				$em = 1 ;
			}
			else
			{
				print "</em>" ;
				$em = 0 ;
			}
		}
		elsif ($tok eq "*")
		{
			if (!$strong)
			{
				print "<strong>" ;
				$strong = 1 ;
			}
			else
			{
				print "</strong>" ;
				$strong = 0 ;
			}
		}
		elsif (isWikiName( $tok ))
		{
			print wikiLinkHtml( $tok ) ;
		}
		elsif (isUrl( $tok ))
		{
			my $url = generaliseUrl( $tok ) ;
			print "<a href=\"$url\">$tok</a>" ;
		}
		else # plain word
		{
			print escape( $tok ) ;
		}
	}
	if ($em)
	{
		print "</em>" ;
	}
	if ($strong)
	{
		print "</strong>" ;
	}
}

# breaks text up into a string of tokens.
# recognises strings of word chars and
# (almost) arbitrary strings delimited by angle brackets as single tokens,
# and calls everything else a delimiter

sub atomise
{
	my $str = shift ;
	my @tokens = () ;
	while (length( $str ) > 0)
	{
		if ($str =~ /^@{[URL_RE]}/) # evilous!
		{
			my $atom = "$1:$2" ;
			push @tokens, $atom ;
			$str = substr( $str, length( $atom )) ;
		}
		elsif ($str =~ /^([A-Za-z0-9]+)/)
		{
			my $atom = $1 ;
			push @tokens, $atom ;
			$str = substr( $str, length( $atom )) ;
		}
		else
		{
			push @tokens, substr( $str, 0, 1 ) ;
			$str = substr( $str, 1 ) ;
		}
	}
	return @tokens ;
}

sub generaliseUrl
{
	my $url = shift ;
	(my ($scheme, $path) = isUrl( $url )) || return viewUrl( BADURL_PAGE ) ;
	$scheme = lc( $scheme ) ;
	if ($scheme eq "isbn")
	{
		return "http://search.barnesandnoble.com/booksearch/isbninquiry.asp?ISBN=" . $path ;
	}
	elsif ($scheme eq "isfdb")
	{
		$path =~ /([A-Za-z]+)\/(.+)/ || return viewUrl( BADURL_PAGE ) ;
		my $class = lc( $1 ) ;
		my $entity = $2 ;
		if ($class eq "author")
		{
			return "http://isfdb.org/cgi-bin/ea.cgi?$entity" ;
		}
		elsif ($class eq "work")
		{
			return "http://isfdb.org/cgi-bin/pw.cgi?$entity" ;
		}
		else
		{
			return viewUrl( BADURL_PAGE ) ;
		}
	}
	elsif ($scheme eq "imdb")
	{
		$path =~ /([A-Za-z]+)\/(.+)/ || return viewUrl( BADURL_PAGE ) ;
		my $class = lc( $1 ) ;
		my $entity = $2 ;
		if ($class eq "title")
		{
			$entity =~ /[0-9]+/ || return viewUrl( BADURL_PAGE ) ;
			return "http://imdb.com/Title?" . $entity ;
		}
		if ($class eq "name")
		{
			# todo: should try to validate entity
			return "http://imdb.com/Name?" . $entity ;
		}
		else
		{
			return viewUrl( BADURL_PAGE ) ;
		}
	}
	elsif ($scheme eq "lj")
	{
		# the LJ sub canonical_username says that a username is a sequence of 1-15 \w chars or hyphens
		if ($path =~ /^([\w\-]{1,15})\/([0-9]+)$/)
		{
			my $user = $1 ;
			my $item = $2 ;
			return "http://www.livejournal.com/talkread.bml?journal=$user&itemid=$item" ;
		}
		elsif ($path =~ /^[\w\-]{1,15}$/)
		{
			return "http://www.livejournal.com/users/$path" ;
		}
		else
		{
			return viewUrl( BADURL_PAGE ) ;
		}
	}
	elsif ($scheme eq "groggs")
	{
		# todo: validate ID
		return "gopher://gopher.publication.org.uk:70/1/rgtp/index%3Fgroggs/" . $path ;
	}
	elsif ($scheme eq "google")
	{
		return "http://www.google.co.uk/search?q=" . $path ;
	}
	elsif ($scheme eq "dmoz")
	{
		$path =~ /^\/?(.*?)\/?$/ ;
		$path = $1 ;
		return "http://dmoz.org/" . $path ;
	}
	elsif ($scheme eq "doi")
	{
		return "http://dx.doi.org/$path" ;
	}
	elsif ($scheme eq "pmid")
	{
		return "http://www.ncbi.nlm.nih.gov/entrez/query.fcgi?cmd=Retrieve&db=PubMed&list_uids=$path&dopt=Abstract" ;
	}
	elsif ($scheme eq "jargon")
	{
		return "http://jargon.publication.org.uk/html/entry/$path.html" ;
	}
	elsif ($scheme eq "rfc")
	{
		$path =~ /\d+/ || return viewUrl( BADURL_PAGE ) ;
		return "ftp://ftp.isi.edu/in-notes/rfc$path.txt" ;
	}
	else
	{
		return $url ;
	}
}

### HTML GENERATION

sub htmlHeader
{
	my $title = shift ;
	my %hdrs = @_ ;
	print <<EOF ;
Content-type: text/html; charset=iso-8859-1
EOF
	$hdrs{"X-Wiki"} = "TwicI" ;
	$hdrs{"X-All-your-base-are-belong-to"} = "us" ;
delete( $hdrs{"Last-Modified"} ) ; # not working too good atm - todo: fix
(!$hdrs{"Last-Modified"}) || die() ;
	foreach my $hdrKey (keys( %hdrs ))
	{
		my $hdrVal = $hdrs{$hdrKey} ;
		print "$hdrKey: $hdrVal\n" ;
	}
	print <<EOF ;

<!doctype HTML PUBLIC "-//W3C//DTD HTML 3.2 Final//EN">
<html>
<head>
<title>$title</title>
</head>
<body>
<h1>$title</h1>
<hr>
EOF
}

sub htmlFooter
{
	my $page = shift ;
	print <<EOF ;
<hr>
<p>
EOF
	if ($page)
	{
		my $editUrl = editUrl( $page ) ;
		my $backlinksUrl = backlinksUrl( $page ) ;
		my $sourceUrl = sourceUrl( $page ) ;
		my $validatorUrl = validatorUrl( $page ) ;
		my $mtimestr = timeToStr( pageMtime( $page ) ) ;
		print <<EOF ;
| <a href="$editUrl">Edit</a>
| <a href="$backlinksUrl">Backlinks</a>
| <a href="$sourceUrl">Source</a>
| <a href="$validatorUrl">Validate</a> 
|
| Last Modified $mtimestr
|
EOF
	}
	foreach my $keyPage (&KEY_PAGES)
	{
		my $keyLinkHtml = wikiLinkHtml( $keyPage ) ;
		print "| $keyLinkHtml\n" ;
	}
	print <<EOF ;
| </p>
</body>
</html>
EOF
}

sub wikiLinkHtml
{
	my $page = shift ;
	my $explodedPage = explodeWikiName( $page ) ;
	if (pageExists( $page ))
	{
		my $url = viewUrl( $page ) ;
		return "<a href=\"$url\">$explodedPage</a>" ;
	}
	else
	{
		my $url = editUrl( $page ) ;
		return "$explodedPage<a href=\"$url\">?</a>" ;
	}
}

sub viewUrl
{
	my $page = shift ;
	return WIKI_BASE . "view.pl?page=" . $page ;
}

sub editUrl
{
	my $page = shift ;
	return WIKI_BASE . "edit.pl?page=" . $page ;
}

sub updateUrl
{
	my $page = shift ;
	return WIKI_BASE . "update.pl?page=" . $page ;
}

sub sourceUrl
{
	my $page = shift ;
	return WIKI_STATIC . $page . ".twici" ;
}

# todo: use relative URLs

sub validatorUrl # todo: use referer-based form, to ease move to relative URLs
{
	my $page = shift ;
	my $viewUrl = viewUrl( $page ) ;
	return "http://validator.w3.org/check?uri=" . $viewUrl ;
}

sub backlinksUrl
{
        my $page = shift ;
        return WIKI_BASE . "backlinks.pl?page=" . $page ;
}

sub changesUrl
{
        return WIKI_BASE . "changes.pl" ;
}

sub viewsUrl
{
        return WIKI_BASE . "views.pl" ;
}

sub escape
{
	my $str = shift ;
	$str =~ s/&/&amp;/g ;
	$str =~ s/</&lt;/g ;
	$str =~ s/>/&gt;/g ;
#	$str =~ s/"/&quot;/g ; # not necessary, and validator chokes
	return $str ;
}

### PAGE LOGIC

sub changes # returns a list of page names, sorted by date
{
	my $days = shift ;
	my $num = shift ;
	# could use opendir, could use glob, could use backticks
	opendir( DIR, WIKI_ROOT ) ;
	my @files = readdir( DIR ) ;
	my @changes = pages() ;
	# todo: more efficient to do this after the sort, binary-chopping to the cutoff and then slicing the list
	if ( $days )
	{
		my $deadline = time() - ($days * 24 * 60 * 60) ;
		@changes = grep {pageMtime( $_ ) ge $deadline} @changes ;
	}
	@changes = sort { pageMtime( $b ) <=> pageMtime( $a ) } @changes ;
	if ( $num )
	{
		my $len = scalar @changes ;
		my $lim = (($num <= $len) ? $num : $len) - 1 ;
		@changes = @changes[0..$lim] ;
	}
	return @changes ;
}

sub views # returns a list of page names, sorted by date
{
	my $days = shift ;
	my $num = shift ;
	# could use opendir, could use glob, could use backticks
	opendir( DIR, WIKI_ROOT ) ;
	my @files = readdir( DIR ) ;
	my @views = pages() ;
	# todo: more efficient to do this after the sort, binary-chopping to the cutoff and then slicing the list
	if ( $days )
	{
		my $deadline = time() - ($days * 24 * 60 * 60) ;
		@views = grep {pageAtime( $_ ) ge $deadline} @views ;
	}
	@views = sort { pageAtime( $b ) <=> pageAtime( $a ) } @views ;
	if ( $num )
	{
		my $len = scalar @views ;
		my $lim = (($num <= $len) ? $num : $len) - 1 ;
		@views = @views[0..$lim] ;
	}
	return @views ;
}

sub backlinks
{
	my $to = shift ;
	my @pages = pages() ;
	my %backlinks = () ;
	foreach my $from (@pages)
	{
		if (isLink( $from, $to ))
		{
			$backlinks{ $from } = undef ;
		}
	}
	return %backlinks ;
}

sub isLink
{
	my $from = shift ;
	my $to = shift ;
	my $fh = openPage( $from ) ;
	my %links = linksInFile( $fh ) ;
	close( $fh ) ;
	return exists( $links{$to} ) ;
}

sub linksInFile # FEEL THE AWESOME POWER OF FULLY-OPERATIONAL LEXICAL CLOSURES!
{
	my $fh = shift ;
	my %links = () ;
	forEachBlockInFile( $fh, sub {
		my $block = shift ;
		my %blockLinks = linksInBlock( $block ) ;
		my @blockLinksLst = keys( %blockLinks ) ;
		foreach my $link (@blockLinksLst)
		{
			$links{$link} = undef ;
		}
	} ) ;
	return %links ;
}

sub forEachBlockInFile # (fh, ref-to-sub( block, context ), context)
{
	my $fh = shift ;
	my $sub = shift ;
	my $oldSlash = $/ ;
	$/ = "" ; # paragraph mode
	while (my $block = <$fh>)
	{
		&$sub( $block ) ;
	}
	$/ = $oldSlash ;
}

sub linksInBlock
{
	my $block = shift ;
	my %links = () ;
	if (isTextBlock( $block ))
	{
		my @atoms = atomise( $block ) ;
		foreach my $atom (@atoms)
		{
			if (isWikiName( $atom ))
			{
				$links{$atom} = undef ;
			}
		}
	}
	return %links ;
}

sub pageExists
{
	my $name = shift ;
	my $file = pageFile( $name ) ;
	return -e $file ;
}

sub pages # list of all page names
{
	opendir( DIR, WIKI_ROOT ) ;
	my @files = readdir( DIR ) ;
	my @pages ;
	foreach my $file (@files)
	{
		if (my $stem = chopExt( $file, WIKI_EXT ))
		{
			push( @pages, validateWikiName( $stem ) ) ;
		}
	}
	closedir( DIR ) ;
	return @pages ;
}

sub pageMtime
{
	my $page = shift ;
	my $file = pageFile( $page ) ;
	return mtime( $file ) ;
}

sub pageAtime
{
	my $page = shift ;
	my $file = pageFile( $page ) ;
	return atime( $file ) ;
}

### UTILITIES

sub timeToStr
{
#      rfc:    Tue, 15 Nov 1994 12:45:26 GMT
	my $time = shift ;
	my @gmtime = gmtime( $time ) ;
	return POSIX::strftime( "%a, %d %b %Y %T GMT", @gmtime ) ;
}

sub factorStrings # "ab", "ac" -> "a", "b", "c"
{
	my $a = shift ;
	my $b = shift ;
	my $len = min( length( $a ), length( $b ) ) ;
	while (substr( $a, 0, $len ) ne substr( $b, 0, $len ))
	{
		--$len ;
	}
	return (substr( $a, 0, $len ), substr( $a, $len ), substr( $b, $len ) ) ;
}

sub min
{
	my $a = shift ;
	my $b = shift ;
	return ($a <= $b) ? $a : $b ;
}

sub chopExt
{
	my $str = shift ;
	my $ext = shift ;
	my $extLen = length( $ext ) ;
	if (substr( $str, -$extLen) eq $ext)
	{
		return substr( $str, 0, -$extLen ) ;
	}
	else
	{
		return undef ;
	}
}

#sub inList
#{
#	my $tgt = shift ;
#	my @lst = @_ ;
#	foreach my $item (@lst)
#	{
#		if ($item eq $tgt)
#		{
#			return 1 ;
#		}
#	}
#	return 0 ;
#}

#sub uniq # uniquify a sorted list
#{
#	my @lst = @_ ;
#	(scalar( @lst ) > 0) || return ;
#	my $last = $lst[0] ;
#	my $j = 1 ;
#	for (my $i = 1 ; $i < scalar( @lst ) ; ++$i)
#	{
#		my $cur = $lst[$i] ;
#		if ($cur ne $last)
#		{
#			$last = $cur ;
#			$lst[$j] = $cur ;
#			++$j ;
#		}
#	}
#	$#lst = ($j - 1) ;
#	return @lst ;
#}

sub normaliseLines
{
	my $text = shift ;
	return join( "\n", split( /\x0D\x0A|\x0A|\x0D/, $text ) ) . "\n" ;
#	$text =~ s/\x0D\x0A|\x0A|\x0D/\n/ ;
		# todo: make this work
	return $text ;
}

sub validateWikiName
{
	my $page = shift ;
	$page || die( "no page name" ) ;
	($page = isWikiName( $page )) || die( "bad page name: $page" ) ;
	return $page ;
}

sub isWikiName
{
	my $str = shift ;
	if ($str =~ /^([A-Z][A-Za-z0-9]*[A-Z][A-Za-z0-9]*)$/)
	{
		return $1 ;
	}
	else
	{
		return undef ;
	}
}

sub explodeWikiName # todo: investigate use of a cunning split() here
{
	my $name = shift ;
#	$name =~ /^([A-Z][A-Za-z0-9]*[A-Z][A-Za-z0-9]*)$/ || die( "not a wiki name: $name" ) ;
	my @atoms = () ;
	my $acc ;
	while ($name ne "")
	{
		if ($name =~ /^([A-Z][a-z]+)(.*)$/)
		{
			if ($acc)
			{
				push( @atoms, $acc ) ;
				$acc = undef ;
			}
			push( @atoms, $1 ) ;
			$name = $2 ;
		}
		else
		{
			my $char = substr( $name, 0, 1 ) ;
			if ($acc)
			{
				$acc .= $char ;
			}
			else
			{
				$acc = $char ;
			}
			$name = substr( $name, 1 ) ;
		}
	}
	if ($acc)
	{
		push( @atoms, $acc ) ;
	}
	return join( " ", @atoms ) ;
}

sub isUrl
{
	my $str = shift ;
	if ($str =~ /^@{[URL_RE]}$/)
	{
		if (wantarray)
		{
			return ($1, $2) ;
		}
		else
		{
			return "$1:$2" ;
		}
	}
	else
	{
		return undef ;
	}
}

sub parse # todo: replace with CGI.pm or CI::Lite call?
{
	my $queryString = shift ;
	my @pairs = split( "&", $queryString ) ;
	my %query ;
	foreach my $pair (@pairs)
	{
	        (my $name, my $value) = split( "=", $pair ) ;
		defined( $value ) || die( "bad pair: $pair" ) ;
	        $value =~ tr/+/ / ;
	        $value =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg ;
	        $name =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg ;
	        $value =~ s/\0//g ;
	        $name  =~ s/\0//g ;
	        $query{$name} = $value ;
	}
	return %query ;
}

### FILE HANDLING

sub pageFile
{
	my $page = shift ;
	return WIKI_ROOT . $page . WIKI_EXT ;
}

sub mtime
{
	my $file = shift ;
	my @stat = stat( $file ) ;
	return $stat[9] ;
}

sub atime
{
	my $file = shift ;
	my @stat = stat( $file ) ;
	return $stat[8] ;
}

sub openPage
{
	my $page = shift ;
	my $mode = shift ;
        my $pageFile = pageFile( $page ) ;
	$mode || ($mode = "<") ;
	fhopen( WIKI_ROOT, $LOCK_HIER{$mode} ) ;
	return fhopen( $pageFile, $mode ) ;
}

# modes:
#  <	read
#  >	write
#  -	access (does not actually open the file!)

sub fhopen # implicitly locks
{
        my $file = shift ;
	my $mode = shift ;
	if (-d $file)
	{
		$file =~ /^(.*?)\/?$/ || die( "invalid filename: $file" ) ;
		$file = $1 ;
		$file = "$file/.dirproxy" ;
	}
	if (!$mode)
	{
		$mode = "<" ;
	}
	($mode ne "-") || return undef ;
	($mode ne "+>") || die( "perverse mode: \'$mode\'" ) ;
	my $fh = fh() ;
	open( $fh, (($mode eq ">") ? ">>" : $mode), $file ) || die( "unable to open file $file in mode like \'$mode\' ($!)" ) ;
	fhlock( $fh, $mode ) ;
	if ($mode eq ">")
	{
		seek( $fh, 0, 0 ) ;
		truncate( $fh, 0 ) ;
	}
	return $fh ;
}

sub fhlock
{
	my $fh = shift ;
	my $mode = shift ;
	my $op ;
	(length( $mode ) > 0) || die( "unspecified mode" ) ;
	$mode = substr( $mode, -1, 1 ) ;
	if ($mode eq "<")
	{
		$op = LOCK_SH ;
	}
	elsif ($mode eq ">")
	{
		$op = LOCK_EX ;
	}
	elsif ($mode eq "-")
	{
		return ;
	}
	else
	{
		die( "unrecognised lock mode \'$mode\'" ) ;
	}
	flock( $fh, $op ) || die( "could not lock filehandle $fh in mode \'$mode\' ($!)" ) ;
}

sub fhclose
{
	my $fh = shift ;
	close( $fh ) || die( "could not close filehandle $fh ($!)" ) ;
}

sub fhunlock
{
	my $fh = shift ;
	flock( $fh, LOCK_UN ) || die( "could not unlock filehandle $fh ($!)" ) ;
}

sub fh
{
	local *FH ;
	return *FH ;
}

### DEBUG

sub STACKTRACE
{
	my $tag = shift ;
	my $depth = shift ;
	$tag || ($tag = "") ;
	$depth || ($depth = 5) ;
	for (my $i = $depth ; $i > 0  ; --$i)
	{
		my ($package, $filename, $line, $subroutine,
			$hasargs, $wantarray, $evaltext, $is_require,
			$hints, $bitmask)
			= caller( $i ) ;
		warn( "$tag: -$i -> $package\:\:$subroutine ($filename $line)" ) ;
	}
}

sub INODE
{
	my $fh = shift ;
	my @stat = stat( $fh ) ;
	my $inode = $stat[1] ;
	return $inode ;
}
